/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     | Website:  https://openfoam.org
    \\  /    A nd           | Copyright (C) 2016-2018 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "RK5.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
namespace RBD
{
namespace rigidBodySolvers
{
    defineTypeNameAndDebug(RK5, 0);
    addToRunTimeSelectionTable(rigidBodySolver, RK5, dictionary);
}
}
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::RBD::rigidBodySolvers::RK5::RK5
(
    rigidBodyMotion& body,
    const dictionary& dict
)
:
    rigidBodySolver(body),
    c2_(0.2),
    c3_(0.3),
    c4_(0.6),
    c5_(1.0),
    c6_(0.875),
    a21_(0.2),
    a31_(3./40.),
    a32_(9./40.),
    a41_(0.3),
    a42_(-0.9),
    a43_(1.2),
    a51_(-11./54.),
    a52_(2.5),
    a53_(-70/27.),
    a54_(35./27.),
    a61_(1631./55296),
    a62_(175./512.),
    a63_(575./13824.),
    a64_(44275./110592.),
    a65_(253./4096.),
    b1_(37./378.),
    b2_(0),
    b3_(250./621.),
    b4_(125./594.),
    b5_(0),
    b6_(512./1771)
{}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::RBD::rigidBodySolvers::RK5::~RK5()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void Foam::RBD::rigidBodySolvers::RK5::solve
(
    const scalarField& tau,
    const Field<spatialVector>& fx
)
{
    Info << "RK5 calc" << endl;
    // Accumulate the restraint forces
    scalarField rtau(tau);
    Field<spatialVector> rfx(fx);
    model_.applyRestraints(rtau, rfx);

    // Calculate the accelerations for the given state and forces
    model_.forwardDynamics(state(), rtau, rfx);

    // Correct velocity
    k1_ = qDdot0();
    k2_ = qDdot0() + deltaT()*(a21_*k1_);
    k3_ = qDdot0() + deltaT()*(a31_*k1_ + a32_*k2_);
    k4_ = qDdot0() + deltaT()*(a41_*k1_ + a42_*k2_ + a43_*k3_);
    k5_ = qDdot0() + deltaT()*(a51_*k1_ + a52_*k2_ + a53_*k3_ + a54_*k4_);

    qDot() = qDot0()
        + deltaT()*(b1_*k1_ + b2_*k2_ + b3_*k3_ + b4_*k4_ + b5_*k5_);

    // Correct position
    k1_ = qDot0();
    k2_ = qDot0() + deltaT()*(a21_*k1_);
    k3_ = qDot0() + deltaT()*(a31_*k1_ + a32_*k2_);
    k4_ = qDot0() + deltaT()*(a41_*k1_ + a42_*k2_ + a43_*k3_);
    k5_ = qDot0() + deltaT()*(a51_*k1_ + a52_*k2_ + a53_*k3_ + a54_*k4_);

    q() = q0()
      + deltaT()*(b1_*k1_ + b2_*k2_ + b3_*k3_ + b4_*k4_ + b5_*k5_);

    correctQuaternionJoints();
}


// ************************************************************************* //
